import{p as e,f as o}from"./flowDb-956e92f1-362cb0e1.js";import{f as t,g as a}from"./styles-c10674c1-784e573a.js";import{u as i}from"./index-e789f855.js";import"./graph-c8c56300.js";import"./layout-60411930.js";import"./index-3862675e-71d2e803.js";import"./clone-55f6f037.js";import"./edges-e0da2a9e-ce972c1a.js";import"./createText-2e5e7dd3-393aac05.js";import"./line-c774d6a4.js";import"./array-9f3ba611.js";import"./path-53f90ab3.js";import"./channel-0e91cabd.js";const M={parser:e,db:o,renderer:t,styles:a,init:r=>{r.flowchart||(r.flowchart={}),r.flowchart.arrowMarkerAbsolute=r.arrowMarkerAbsolute,i({flowchart:{arrowMarkerAbsolute:r.arrowMarkerAbsolute}}),t.setConf(r.flowchart),o.clear(),o.setGen("gen-2")}};export{M as diagram};
