export const rasaZIndices = {
  hide: -1,
  auto: "auto",
  base: 0,
  docked: 1,
  dropdown: 2,
  sticky: 3,
  banner: 4,
  overlay: 5,
  popover: 7,
  skipLink: 8,
  toast: 9,
  tooltip: 10,
  modalOverlay: 100,
  modal: 101,
  alwaysOnTop: 9999,
};

export type RasaZIndices = typeof rasaZIndices;
